
### EVALUATOR FOR LAMBDA EXPRESSIONS extended with Python code fragments
###  for imitating delta-operators

import OperatorTable  # holds definitions of delta-operators
import Format         # holds functions for conversions between strings and ETREEs
                      #   and function for writing to output log file

"""Input format:

ETREE ::=  [ ETREE+ ]  |  [ "lam", ID, ETREE ]  |  ID
        |  ( "code", PRINTNAME, ARITY, PYTHONCODE )

"""

def eval(e) :
    """pre:  e is an ETREE
       post: computes a sequence of leftmost outermost reduction steps on e,
             printing them on display and logging them in filename
    """
    print "Press Enter to single-step the reduction; press  n  to quit now; press  y  to reduce to normal form; press Ctrl+c to interrupt:"
    print
    goToTheEnd = False
    stage = e           # current stage of reduced  e
    step = e            # outcome of most recent reduction step
    while step != 0 :   #  0  means no more reductions required
        #print "stage:", stage  ####### UNCOMMENT TO SEE RAW VALUE
        stringStage = Format.toString(stage)
        Format.writeLine(stringStage)
        if goToTheEnd :
            import time
            time.sleep(0.3) # delays for 0.3 second
            step = eval1(stage)
        else :
            user = raw_input()
            if user == 'n' :  step = 0
            else :     
                if user == 'y' : goToTheEnd = True       
                step = eval1(stage)
        if step != 0 :
            stage = step
    Format.writeLine("FINISHED.")
            

def eval1(e) :  
    """pre:  e : ETREE
       returns:  e' : ETREE  or  0
       post:  0 means no reduction step taken;   
              e'  means  e => e'  by a single leftmost reduction step
    """
    if isinstance(e, str) :      # ID
        if e in OperatorTable.ops :
            return OperatorTable.ops[e]
        else :  return 0
    elif isinstance(e, tuple) :  # TUPLE
        return 0  # operator is in normal form
    else :
        op = e[0]
        if op == "lam" :   
            e2 = eval1(e[2])     # [ "lam", ID, ETREE ] 
            if e2 == 0 : return 0
            else :  return ["lam", e[1], e2]
        else :                   #  [ ETREE+ ]
            if len(e) == 1 :     #  [ ETREE ]  with superfluous brackets
                return e[0]
            # else :   # [ ETREE0, ETREE1, ... ]
            if isLambdaAbs(e[0]) :
                return doBetaReduction(e)
            # else : we have an application that must be further reduced:
            index = 0
            for Ei in e :
                # finding a leftmost reduction in the application:
                ei = eval1(Ei)
                if ei != 0 : break
                else :  index = index + 1
            if ei == 0 : 
                # all args in application are in normal form.
                # can we do a delta step?
                if isinstance(e[0], tuple) and e[0][0] == "code" : 
                    return tryDeltaReduction(e)
                else : return 0
            else :
                return e[:index] + [ei] + e[index+1:]

def tryDeltaReduction(e) :
    """pre: e  has form  [ ("code", name, arity, pythoncode), E1, ..., Ek ]
            where all Ei are in normal form 
       post:  applies first  arity-many  Ei  arguments to pythoncode
              and assembles the answer
    """
    (tag, name, arity, pycode) = e[0]
    if arity > 0 and arity <= len(e[1:]) :   
        # have enough args to do delta-step: assemble the args and call:
        args = ()
        for i in range(arity) :
            args = args + (e[i+1],)
        answer = pycode(args)
        suffix = e[arity+1:]
        if suffix != [] :
              return  [answer] + e[arity+1:]
        else :
              return  answer
    else :  return 0
        


def isLambdaAbs(e) :
    """returns whether  e  has form  ["lam", ID, ETREE]"""
    return  isinstance(e, list) and e[0] == "lam"

def doBetaReduction(e) :
    """pre: e  has form  [ ["lam", I, E0], E1, ..., Ek ]
       returns:  e'
       post:  e => e'  by beta reduction, when k > 0
    """
    contractum = subst(e[1], e[0][1], e[0][2])
    #print "contractum", contractum
    if len(e) > 2 :  # multiple args in this application?
        return [contractum] + e[2:]
    else :  # application is all used up!
        return contractum

def subst(arg, id, body) :
    """pre:  arg is ETREE,  id is ID,  body is ETREE
       post: returns  [arg/id]body,  the usual painful stuff
    """
    if isinstance(body, str) :     # ID
        if body == id : return arg
        else : return body
    elif isinstance(body, tuple) :  # TUPLE
        return body   # a tuple is a constant atom
    else :
        op = body[0]
        if op == "lam" :       # [ "lam", ID, ETREE ]
            binderid = body[1]
            if id == binderid :
                return body
            elif not(freein(binderid, arg)) :
                return ["lam", binderid, subst(arg, id, body[2])]
            else :
                newbinder = freshname(binderid)
                newbody = subst(newbinder, binderid, body[2])
                return ["lam", newbinder, subst(arg, id, newbody)]
        else :                   #  [  ELIST ]
            ans = []
            for ei in body :
                ans.append(subst(arg, id, ei))
            return ans
            
def freein(id, e) :
    """returns whether  id  appears unbound in expr  e"""
    if isinstance(e, str) :     # ID
        return id == e
    elif isinstance(e, tuple) :  # TUPLE
        return False
    else :
        op = e[0]
        if op == "lam" :       # [ "lam", ID, ETREE ]
            if id == e[1] : return False
            else : return freein(id, e[2])
        else:         #  [ ELIST ]
            ans = False
            for ei in e :
                 ans = ans or freein(id, ei)
            return ans

namecount = 0
def freshname(id) :
    """returns a new  id  that is different from  id"""
    global namecount
    newid = id + str(namecount)
    namecount = namecount + 1
    return newid
