
### Format.py  reads the input file, formats it, and generates the output,
###    archiving it in a text file.


import OperatorTable   # table of delta ops

"""

Input format:  a string, written like ETREE, but IDs are not quoted yet.

Expression format:
ETREE ::=  [ ETREE+ ]  |  [ "lam", ID, ETREE ]  |  ID
        |  ( "code", PRINTNAME, ARITY, PYTHONCODE )
ID ::=  alphanumeric string

"""

def toETREE(s) :
    """pre:  s is a string in input format
       post: returned answer is a Python ETREE expression,
              assembled by quoting all IDs  and calling  eval
    """
    etree = "ERROR0"
    formatted = ""
    quoteMode = False
    index = 0
    for c in s :
        if c.isalnum() :
            if not quoteMode:
                formatted = formatted + "'"
                quoteMode = True
        elif quoteMode :
            formatted = formatted + "'"
            quoteMode = False
        formatted = formatted + c
        index = index + 1
    if quoteMode :
        formatted = formatted + "'"
    try :
        etree = eval(formatted)
    except:  # an error:
       import traceback
       traceback.print_exc()
       print
       writeLine("FORMAT ERROR - expression is malformed:\n" + s)
       print
       etree = "ERROR1"
    return etree


def toString(e) :
    """returns a prettier version of ETREE e for printing"""
    if isinstance(e, str) :      # ID
        return e
    elif isinstance(e, tuple) :  # TUPLE
        if e[0] == "code" :
            #return str(e[3]) ######### RETURNS INTERNAL PYTHON VALUE OF code
            if e[1] != "" :  return "'"+e[1]    # the code has a useful name
            else :           return str(e[3])   # else, return code itself
        else : return str(e)
    else :
        op = e[0]
        if op == "lam" :       # [ "lam", ID, ETREE ]
            return "[lam " + e[1] + ". " + toString(e[2]) + "]"  
        else :                 #  [ ETREE+ ]
            ans = "["
            for ei in e :
                ans = ans + toString(ei) + " "
            return ans + "]"

            
            
#### files from where to read input and dump output:           
inputfile = ""
outputfile = ""

def initialize(fname = "") :
    """ 1. opens file  fname, if any,  and its log file,  fnameLog.txt
        2. reads def-lines from  fname,  each line has form,  def NAME : EXPR
           and converts EXPR to an ETREE and adds  {NAME : ETREE} to operator table
        3. reads expression to be reduced, if any, converts it to an ETREE, 
           and returns it  (returns "" if no expression to be reduced in fname)
    """
    global inputfile, outputfile
    
    answer = ""
    log = "Log.txt"        # name of output file
    OperatorTable.reset()

    if fname != "" :
        dot = fname.find(".")
        if dot != -1 :
           log = fname[:dot] + "Log.txt"
        else :
           log = fname + "Log.txt" 

    outputfile = open(log, "w")
    if fname == "" : return answer  
    
    # else: there is an input file to read....

    # read input file for defs and possible expr to reduce:
    inputfile = open(fname, "r")
    line = inputfile.readline()
    while line != "" :        # while not EOF:
        line = line.strip()   #   remove leading/trailing white space
        
        if line == "" :       #   blank line; ignore it
            line = inputfile.readline()        
            continue   
        
        if len(line) >= 4  and  line[:4] == "def " :  #  def NAME : EXPR
            line1 = line[4:]
            colon = line1.find(":")
            if colon == -1 :
                message = "IGNORE LINE - def is missing colon: " + line
            else :
                # assemble, format, and save the def:
                namepart = line1[:colon].strip()
                exprpart = line1[colon+1:].strip()
                expr = toETREE(exprpart)
                outcome = OperatorTable.addOp(namepart, expr)
                message = "def " + namepart + " : " + str(expr)
                if outcome :
                    message = "OK " + message
                else :
                    message = "ERROR " + message
            writeLine(message)
            line = inputfile.readline()
            
        else :  # there is an expression to reduce:
            answer = toETREE(line)
            writeLine("reduce this expression: " + str(answer))
            break     # done --- we return the first expr we see....
    
    writeLine("END READING INPUT FILE")
    return answer    

  
def writeLine(text) :
    """writes text line to output file"""
    print text
    outputfile.write(text + "\r\n")    


def finalize():
    if inputfile != "" : inputfile.close()
    outputfile.close()


def deleteFile(filename) :
    """deletes the  filename  from the folder"""
    import os
    os.remove(filename)


