
"""operator table of user-defined names and their definitions

   Format of bindings:
     name : VAL
               where VAL is a closed lambda expression
               or  ("code", PRINTNAME, ARITY, pythonCode)
"""

####### User-defined functions go here.  The table is at the bottom of the file:

# Environment table modelled as a dictionary of name:value bindings:

recursionAllowed = False    # whether we allow direct self-reference, e.g.,
                            # def loop : [lam, x, [loop, x]]  is a self-reference
initenv = {}

def find((id, (codetag, name, arity, env))) :
     if codetag == "code" :
         return env[id]
     else : return "ERROR"

def bind((id, value, (codetag, name, arity, env))) :
     if codetag == "code" :
         env[id] = value
         return ("code", "",  arity, env)
     else : return "ERROR"

#########



###### HERE IS THE TABLE: ######

ops = { 

"zero" : ["lam", "s", ["lam", "z", "z"]],

"one" : ["lam", "s", ["lam", "z", ["s", "z"]]],

"two" : ["lam", "s", ["lam", "z", ["s", ["s", "z"]]]],

"three" : ["lam", "s", ["lam", "z", ["s", ["s", ["s", "z"]]]]],

"True" : ["lam", "t", ["lam", "e", "t"]],

"False" : ["lam", "t", ["lam", "e", "e"]],

"not" : ["lam", "b", ["b", "False", "True"]],

"eq0" : ["lam", "m", ["m", ["lam", "z", "False"], "True"]],

"succ" : ["lam", "n", ["lam", "s", ["lam", "z", ["s", ["n", "s", "z"]]]]],

"add" : ["lam", "m", ["lam", "n", ["m", "succ", "n"]]],


"initenv": ("code", "initenv", 0, initenv),

"find":  ("code", "find", 2,  find),

"bind":  ("code", "bind", 3, bind),

}

# function to add an operator to the table:
#   name is a string,  val is a *closed* lambda expression
#   returns whether the addition was successful 
def addOp(name, value):

    def FreeVars(e) :  # returns list of the free vars in expr e
        if isinstance(e, str) :     # ID
            return [e]
        elif isinstance(e, tuple) :  # TUPLE
            return []
        else :
            op = e[0]
            if op == "lam" :       # [ "lam", ID, ETREE ]
                id = e[1]
                return [i for i in FreeVars(e[2]) if i != id]
                #remove(id, FreeVars(e[2])
            else:  #  [ ELIST ]
                ans = []
                for ei in e :
                     ans = ans + FreeVars(ei)
                return ans
    # end  
    
    ids = FreeVars(value)
    for id in ids :
        if id != name and id not in ops :
            return   # free var error
    if name in ids and not recursionAllowed : return   # recursion error
    ops[name] = value
    return True

# function to erase the table  )-:    
def reset() :
    global ops
    ops = {}


